@@ -4,6 +4,7 @@ import torch
 import pyro
 from pyro.infer import SVI,Trace_ELBO,JitTrace_ELBO
 from pyro.optim import Adam
+import pyro.distributions.constraints as constraints
 import pyro.distributions as dist
 import torch.nn.functional as F
 from tqdm import trange
@@ -13,7 +14,7 @@ from tqdm import trange
 
 class PyBasilica():
 
-    def __init__(self, x, k_denovo, lr, n_steps, groups=None, beta_fixed=None, compile_model = True, CUDA = False):
+    def __init__(self, x, k_denovo, lr, n_steps, groups=None, beta_fixed=None, compile_model = True, CUDA = False, enforce_sparsity = False):
         self._set_data_catalogue(x)
         self._set_beta_fixed(beta_fixed)
         self.k_denovo = int(k_denovo)
@@ -21,6 +22,7 @@ class PyBasilica():
         self.n_steps = int(n_steps)
         self.compile_model = compile_model  
         self.CUDA = CUDA
+        self.enforce_sparsity = enforce_sparsity
         self._set_groups(groups)
         self._check_args()
     
@@ -83,34 +85,42 @@ class PyBasilica():
 
             with pyro.plate("k", k_fixed + k_denovo):   # columns
                 with pyro.plate("n", n_samples):        # rows
-                    alpha = pyro.sample("latent_exposure", dist.Normal(alpha_mean, 1))
+                    if self.enforce_sparsity:
+                        alpha = pyro.sample("latent_exposure", dist.Exponential(3))
+                    else:
+                        alpha = pyro.sample("latent_exposure", dist.HalfNormal(1))
         
-        alpha = torch.exp(alpha)                                # enforce non negativity
         alpha = alpha / (torch.sum(alpha, 1).unsqueeze(-1))     # normalize
+        alpha = torch.clamp(alpha, 0,1)
 
         #----------------------------- [BETA] -------------------------------------
         if k_denovo==0:
             beta_denovo = None
         else:
-            beta_mean = dist.Normal(torch.zeros(k_denovo, 96), 1).sample()
+            #beta_mean = dist.Normal(torch.zeros(k_denovo, 96), 1).sample()
             with pyro.plate("contexts", 96):            # columns
                 with pyro.plate("k_denovo", k_denovo):  # rows
-                    beta_denovo = pyro.sample("latent_signatures", dist.Normal(beta_mean, 1))
-            beta_denovo = torch.exp(beta_denovo)                                    # enforce non negativity
+                    beta_denovo = pyro.sample("latent_signatures", dist.HalfNormal(1))
             beta_denovo = beta_denovo / (torch.sum(beta_denovo, 1).unsqueeze(-1))   # normalize
+            beta_denovo = torch.clamp(beta_denovo, 0,1)
 
         #----------------------------- [LIKELIHOOD] -------------------------------------
         if self.beta_fixed is None:
             beta = beta_denovo
+            reg = 0
         elif beta_denovo is None:
             beta = self.beta_fixed
+            reg = 0
         else:
             beta = torch.cat((self.beta_fixed, beta_denovo), axis=0)
+            reg = self._regularizer(self.beta_fixed, beta_denovo)
+        
+        
         
         with pyro.plate("contexts2", 96):
             with pyro.plate("n2", n_samples):
-                pyro.sample("obs", dist.Poisson(torch.matmul(torch.matmul(torch.diag(torch.sum(self.x, axis=1)), alpha), beta)), obs=self.x)
-                
+                lk =  dist.Poisson(torch.matmul(torch.matmul(torch.diag(torch.sum(self.x, axis=1)), alpha), beta)).log_prob(self.x)
+                pyro.factor("loss", lk - reg)
     
 
     def guide(self):
@@ -120,23 +130,23 @@ class PyBasilica():
         k_denovo = self.k_denovo
         #groups = self.groups
 
-        alpha_mean = dist.Normal(torch.zeros(n_samples, k_fixed + k_denovo), 1).sample()
+        alpha_mean = dist.HalfNormal(torch.ones(n_samples, k_fixed + k_denovo)).sample()
 
         with pyro.plate("k", k_fixed + k_denovo):
             with pyro.plate("n", n_samples):
-                alpha = pyro.param("alpha", alpha_mean)
+                alpha = pyro.param("alpha", alpha_mean, constraint=constraints.greater_than_eq(0))
                 pyro.sample("latent_exposure", dist.Delta(alpha))
 
         if k_denovo != 0:
-            beta_mean = dist.Normal(torch.zeros(k_denovo, 96), 1).sample()
+            beta_mean = dist.HalfNormal(torch.ones(k_denovo, 96)).sample()
             with pyro.plate("contexts", 96):
                 with pyro.plate("k_denovo", k_denovo):
-                    beta = pyro.param("beta_denovo", beta_mean)
+                    beta = pyro.param("beta_denovo", beta_mean, constraint=constraints.greater_than_eq(0))
                     pyro.sample("latent_signatures", dist.Delta(beta))
 
-    '''
+    
     def _regularizer(self, beta_fixed, beta_denovo):
-
+        '''
         if beta_denovo == None:
             dd = 0
         else:
@@ -149,19 +159,15 @@ class PyBasilica():
                     c2 += 1
                     if c1!=c2:
                         dd += F.kl_div(denovo1, denovo2, reduction="batchmean").item()
-
-        if beta_fixed == None or beta_denovo == None:
-            loss = 0
-        else:
-            #cosi = torch.nn.CosineSimilarity(dim=0)
-            loss = 0
-            for fixed in beta_fixed:
-                for denovo in beta_denovo:
-                    loss += F.kl_div(fixed, denovo, reduction="batchmean").item()
-                    #loss += cosi(fixed, denovo).item()
-            #print("loss:", loss)
-        return loss + (dd/2)
-    '''
+        '''
+        loss = 0
+        for fixed in beta_fixed:
+            for denovo in beta_denovo:
+                loss += F.kl_div(torch.log(fixed), torch.log(denovo), log_target = True, reduction="batchmean")
+                #loss += cosi(fixed, denovo).item()
+        #print("loss:", loss)
+        return loss
+    
     
     def _likelihood(self, M, alpha, beta_fixed, beta_denovo):
         
@@ -215,14 +221,14 @@ class PyBasilica():
 
             # create likelihoods -------------------------------------------------------------
             alpha = pyro.param("alpha").clone().detach()
-            alpha = torch.exp(alpha)
+            #alpha = torch.exp(alpha)
             alpha = alpha / (torch.sum(alpha, 1).unsqueeze(-1))
 
             if self.k_denovo == 0:
                 beta_denovo = None
             else:
                 beta_denovo = pyro.param("beta_denovo").clone().detach()
-                beta_denovo = torch.exp(beta_denovo)
+                #beta_denovo = torch.exp(beta_denovo)
                 beta_denovo = beta_denovo / (torch.sum(beta_denovo, 1).unsqueeze(-1))
 
             likelihoods.append(self._likelihood(self.x, alpha, self.beta_fixed, beta_denovo))
@@ -267,7 +273,7 @@ class PyBasilica():
         if self.CUDA and torch.cuda.is_available():
           alpha = alpha.cpu()
         alpha = alpha.clone().detach()
-        alpha = torch.exp(alpha)
+        #alpha = torch.exp(alpha)
         self.alpha = alpha / (torch.sum(alpha, 1).unsqueeze(-1))
 
     
@@ -280,7 +286,7 @@ class PyBasilica():
             if self.CUDA and torch.cuda.is_available():
               beta_denovo = beta_denovo.cpu()
             beta_denovo = beta_denovo.clone().detach()
-            beta_denovo = torch.exp(beta_denovo)
+            #beta_denovo = torch.exp(beta_denovo)
             self.beta_denovo = beta_denovo / (torch.sum(beta_denovo, 1).unsqueeze(-1))
     
 
@@ -358,6 +364,7 @@ import warnings
 warnings.filterwarnings('ignore')
 
 def is_stationary(data: pd.Series, alpha: float = 0.05):
+  
     # Test to see if the time series is already stationary
     if kpss(data, regression='c', nlags="auto")[1] > alpha:
     #if adfuller(data)[1] < alpha:
@@ -368,9 +375,11 @@ def is_stationary(data: pd.Series, alpha: float = 0.05):
         return False
 
 def convergence(x, alpha: float = 0.05):
+  ### !!! REMEMBER TO CHECK !!! ###
+    return False
     if isinstance(x, list):
         data = pd.Series(x)
     else:
         raise Exception("input list is not valid type!, expected list.")
 
-    return is_stationary(data, alpha=alpha)
\ No newline at end of file
+    return is_stationary(data, alpha=alpha)
